import ast

from .base_pass import BasePass
from ..visitors import CollectLineNumbers
from ..common import a2s, COMMENT_MARKER


class DeadcodePass(BasePass):
    """
    Eliminated un-executed code paths.

    Dead code elimination uses a runtime trace to determine whether a node
    in the AST was actually executed. For example, if an if/else only ever
    visits the else branch, we can eliminate the if branch.

    To determine whether code was executed, the Python tracing facilities
    tell us whether a line of code in the source file was executed during a
    trace. For a given syntax object, we can compute the set of line numbers
    it spans, and check if any of the lines were executed.

    Example:
      flag = False
      if flag:
        x = 1
      else:
        x = 2

      >> becomes >>

      flag = False
      x = 2
    """
    tracer_args = {'trace_opcodes': True, 'trace_lines': True}

    def _is_comment(self, node):
        return isinstance(node, ast.Expr) and \
            isinstance(node.value, ast.Str) and \
            node.value.s.startswith(COMMENT_MARKER)

    def _len_without_comment(self, stmts):
        # We never treat comments as dead code, so a dead block can end up
        # with only comments. Hence we need to count statements while
        # excluding comments.
        return len([s for s in stmts if not self._is_comment(s)])

    def _is_dead(self, node):
        collect_lineno = CollectLineNumbers()
        collect_lineno.visit(node)
        return sum(
            [self.tracer.execed_lines[i] for i in collect_lineno.linenos]) == 0

    def generic_visit(self, node):
        # Don't delete comments
        if self._is_comment(node):
            return node

        # Delete assignments and expression statements that are dead
        if isinstance(node, (ast.Assign, ast.Expr)) and self._is_dead(node):
            self.change = True
            return None

        return super().generic_visit(node)

    def visit_If(self, stmt):
        # TODO: assumes pure conditions
        if self._len_without_comment(stmt.body) == 0 or self._is_dead(
                stmt.body[0]):
            self.change = True
            return stmt.orelse
        elif self._len_without_comment(stmt.orelse) == 0 or self._is_dead(
                stmt.orelse[0]):
            self.change = True
            return stmt.body

        self.generic_visit(stmt)
        return stmt

    def visit_For(self, stmt):
        if self._len_without_comment(stmt.body) == 0:
            self.change = True
            return None

        self.generic_visit(stmt)
        return stmt

    def visit_Expr(self, stmt):
        if self._is_comment(stmt):
            return stmt

        if isinstance(stmt.value, (ast.Name, ast.Str, ast.NameConstant)):
            self.change = True
            return None

        self.generic_visit(stmt)
        return stmt

    def visit_Try(self, stmt):
        if not self._is_dead(stmt.handlers[0]):
            # HUGE HACK: assumes it's safe to replace try/except
            # with just except block if except block not dead
            assert self._len_without_comment(stmt.handlers) == 1
            assert stmt.handlers[0].name is None
            self.change = True
            return stmt.handlers[0].body
        else:
            self.change = True
            return stmt.body



import ast
import inspect

from ..common import make_name, SEP, compare_ast, parse_expr
from .base_pass import BasePass


class FindObjNew(ast.NodeVisitor):
    def __init__(self, globls):
        self.globls = globls
        self.objs = set()

    def visit_Assign(self, stmt):
        if isinstance(stmt.targets[0], ast.Name):
            name = stmt.targets[0].id
            assert name in self.globls, f'{name} not in globals'
            obj = self.globls[name]

            cls = obj.__class__.__name__
            if compare_ast(stmt.value, parse_expr(f'{cls}.__new__({cls})')):
                self.objs.add(id(obj))

    def visit_FunctionDef(self, fdef):
        pass


class UnsafeToExpand(ast.NodeVisitor):
    def __init__(self, globls):
        self.unsafe = set()
        self.globls = globls

    def visit_Call(self, call):
        for arg in call.args:
            if isinstance(arg, ast.Name):
                self.unsafe.add(arg.id)

        self.generic_visit(call)


class ExpandSelfPass(BasePass):
    """
    Replaces all class instances with variables.

    Once an object's methods have been fully inlined, the object is essentially
    a fancy dictionary. We can inline the dictionary by converting each key
    to a unique variable name. Then we can replace object assignments by copying
    every variable individually.

    See ContextualTransforms.inline_constructor for how objects are inlined
    in the first place.

    Example:
      foo = Foo.__new__(Foo)
      foo.x = 1
      bar = foo
      assert bar.x == 1

      >> becomes >>

      foo_x = 1
      bar_x = foo_x
      assert bar_x == 1
    """

    tracer_args = {}

    def visit_Module(self, mod):
        unsafe = UnsafeToExpand(self.globls)
        unsafe.visit(mod)

        finder = FindObjNew(self.globls)
        finder.visit(mod)

        self.objs_to_inline = {}
        # We find all the objects that need to be inlined by going through
        # the globals of the last trace
        for var, obj in self.globls.items():

            # There is no inspect.isobject or inspect.iscreatedfromclass
            # unfortunately. So we proceed by process of elimination. If
            # an object is neither a class or a module, it must be an object
            # so we register it to be inlined.
            if self.inliner.should_inline(obj) and \
               not inspect.isclass(obj) and \
               not inspect.ismodule(obj) and \
               not var in unsafe.unsafe:

                if id(obj) not in self.objs_to_inline and id(
                        obj) in finder.objs:

                    # We heuristically devise a name for the object
                    if hasattr(obj, '__name__'):
                        name = obj.__name__
                    elif hasattr(obj, '__class__'):
                        name = obj.__class__.__name__
                    else:
                        name = 'var'
                    self.objs_to_inline[id(obj)] = self.inliner.fresh(
                        name.lower())

        return super().visit_Module(mod)

    def visit_Attribute(self, attr):
        if isinstance(attr.value, ast.Name):
            name = attr.value.id
            if name in self.globls:
                obj = self.globls[name]
                if id(obj) in self.objs_to_inline:
                    new_name = self.objs_to_inline[id(obj)]
                    self.change = True
                    return make_name(f'{attr.attr}{SEP}{new_name}')

        self.generic_visit(attr)
        return attr

    def visit_Assign(self, stmt):
        if isinstance(stmt.targets[0], ast.Name):
            name = stmt.targets[0].id
            assert name in self.globls
            obj = self.globls[name]

            if id(obj) in self.objs_to_inline:
                new_name = self.objs_to_inline[id(obj)]

                cls = obj.__class__.__name__
                if compare_ast(stmt.value, parse_expr(f'{cls}.__new__({cls})')):
                    self.change = True
                    return [
                        ast.Assign(targets=[make_name(f'{new_name}{SEP}{k}')],
                                   value=ast.NameConstant(None))
                        for k in vars(obj).keys()
                    ]
                else:
                    return []

        self.generic_visit(stmt)
        return stmt



import ast
import copy
from iterextras import unzip

from .base_pass import BasePass
from ..common import a2s, parse_stmt, make_name, SEP


class SimplifyVarargsPass(BasePass):
    """
    Expand/inline usage of *args and **kwargs.

    If a dictionary is being used as keyword arguments to a function,
    this pass attempts to turn each key of the dictionary into a standalone
    variable so it can be optimized by e.g. copy propagation.

    Example:
      kwargs = {'x': 1}
      foo(**kwargs)

      >> becomes >>

      foo(x=1)
    """

    tracer_args = {}

    def _is_vararg(self, node):
        # HEURISTIC: assume that all dictionaries starting with "args"
        # or "kwargs" are used for keyword arguments
        return isinstance(node, ast.Name) and \
            (node.id.startswith('args') or node.id.startswith('kwargs'))

    def visit_Expr(self, stmt):
        # find expressions like dict.update(...)
        if isinstance(stmt.value, ast.Call) and \
           isinstance(stmt.value.func, ast.Attribute) and \
           len(stmt.value.args) > 0 and \
           stmt.value.func.attr == 'update':

            try:
                dict_obj = eval(a2s(stmt.value.func.value), self.globls,
                                self.globls)
            except Exception:
                raise

            # Expand dict.update(x=2) into dict['x'] = 2
            dict_ast = stmt.value.func.value
            arg = stmt.value.args[0]
            if isinstance(dict_obj, dict) and \
               isinstance(arg, ast.Call) and \
               isinstance(arg.func, ast.Name) and \
               arg.func.id == 'dict':
                new_stmts = []
                for kw in arg.keywords:
                    assgn = parse_stmt(f'x["{kw.arg}"] = x')
                    assgn.targets[0].value = dict_ast
                    assgn.value = kw.value
                    new_stmts.append(assgn)

                self.change = True

                # Recursively visit the expansion to continue turning
                # dictionary into variables
                return [self.visit(stmt) for stmt in new_stmts]

        self.generic_visit(stmt)
        return stmt

    def visit_Assign(self, stmt):
        if self._is_vararg(stmt.targets[0]):
            # x = dict(foo='bar', ...)
            if isinstance(stmt.value, ast.Call) and \
               isinstance(stmt.value.func, ast.Name) and \
               stmt.value.func.id == 'dict':
                keys, vals = unzip([(kw.arg, kw.value)
                                    for kw in stmt.value.keywords])
                stmt.value = ast.Dict([ast.Str(k) for k in keys], vals)
                self.change = True
                return self.visit(stmt)

            # x = {'foo': 'bar', ...}
            elif isinstance(stmt.value, ast.Dict) and len(stmt.value.keys) > 0:
                name = stmt.targets[0].id
                new_stmts = [parse_stmt(f'{name} = {{}}')]
                for k, v in zip(stmt.value.keys, stmt.value.values):
                    assgn = parse_stmt(f'{name}[_] = _')
                    assgn.targets[0].slice.value = k
                    assgn.value = v
                    new_stmts.append(assgn)

                self.change = True
                return [self.visit(stmt) for stmt in new_stmts]

        self.generic_visit(stmt)
        return stmt

    def visit_Subscript(self, sub):
        # x['foo'] => foo___x
        if self._is_vararg(sub.value) and \
           isinstance(sub.slice, ast.Index) and \
           isinstance(sub.slice.value, ast.Str):
            return make_name(f'{sub.slice.value.s}{SEP}{sub.value.id}')

        self.generic_visit(sub)
        return sub

    def visit_Call(self, call):
        if len(call.args) > 0 and isinstance(call.args[-1], ast.Starred):
            star_arg = call.args[-1].value

            try:
                star_obj = eval(a2s(star_arg), self.globls, self.globls)
            except Exception:
                # print('ERROR', a2s(call))
                # TODO: log a warning?
                # list comprehension issue: [f(*c) for c in whatever]
                return call

            call.args.pop()
            self.change = True

            for i in range(len(star_obj)):
                call.args.append(ast.Subscript(value=star_arg, slice=ast.Index(ast.Num(i))))

        kwarg = [(i, kw.value) for i, kw in enumerate(call.keywords)
                 if kw.arg is None]

        if len(kwarg) == 1:
            i, kwarg = kwarg[0]

            try:
                kwarg_obj = eval(a2s(kwarg), self.globls, self.globls)
            except Exception:
                print('ERROR', a2s(call))
                raise

            self.change = True
            del call.keywords[i]

            # Add a keyword argument for each key in kwargs, e.g.
            # kwargs = {'x': 1}; f(**kwargs) => f(x=kwargs['x'])
            for k in kwarg_obj.keys():
                idx_expr = ast.Subscript(value=kwarg,
                                         slice=ast.Index(ast.Str(k)))
                call.keywords.append(ast.keyword(arg=k, value=idx_expr))

        self.generic_visit(call)
        return call



import ast

from .base_pass import BasePass
from ..common import SEP, parse_expr, COMMENT_MARKER, a2s
from collections import defaultdict


class RemoveSuffixesPass(BasePass):
    """
    Remove the debug suffixes applied during inlining to avoid name clashes.

    Example:
      x__foo = 1
      x__bar = 2
      assert x__foo + x__bar == 3

      >> becomes >>

      x = 1
      x_2 = 2
      assert x + x_2 == 3
    """
    def __init__(self, inliner):
        super().__init__(inliner)
        self.name_map = {}

    def visit_Name(self, name):
        parts = name.id.split(SEP)
        if len(parts) > 1:
            base = parts[0]
            if name.id not in self.name_map:
                self.name_map[name.id] = self.inliner.fresh(base)
            name.id = self.name_map[name.id]
        return name

    def visit_Expr(self, expr):
        if isinstance(expr.value, ast.Str) and \
           expr.value.s.startswith(COMMENT_MARKER):
            comment = expr.value.s[len(COMMENT_MARKER):]
            call = parse_expr(comment)
            name_map = self.name_map.copy()
            generated_vars = self.inliner.generated_vars.copy()
            self.name_map = {}
            self.inliner.generated_vars = defaultdict(int)
            self.visit(call)
            self.inliner.generated_vars = generated_vars
            self.name_map = name_map
            expr.value.s = COMMENT_MARKER + a2s(call)
            return expr

        self.generic_visit(expr)
        return expr



import ast
from .base_pass import BasePass
from ..common import a2s


class ArrayIndexPass(BasePass):
    """
    Inline indexes into array literals.

    Example:
      x = [1, 2]
      y = x[0]

      >> becomes >>

      y = 1

    Example:
      y = [1, 2][0]

      >> becomes >>

      y = 1
    """
    def __init__(self, inliner):
        super().__init__(inliner)

        # Mapping from names to array literals
        self.arrays = {}

        # Mapping from names to dict literals
        self.dicts = {}

    def visit_Assign(self, assgn):
        super().generic_visit(assgn)

        if len(assgn.targets) == 1 and isinstance(assgn.targets[0], ast.Name):
            # If assigning to an array literal, record the name/value in self.arrays
            name = assgn.targets[0].id
            if isinstance(assgn.value, ast.List):
                self.arrays[name] = assgn.value
            elif isinstance(assgn.value, ast.Dict):
                self.dicts[name] = assgn.value

        return assgn

    def _dict_lookup(self, dct, key):
        return {k.s: v for k, v in zip(dct.keys, dct.values)}[key]

    def visit_Subscript(self, expr):
        # If indexing with a constant value, then get the corresponding element
        if isinstance(expr.slice, ast.Index):
            if isinstance(expr.slice.value, ast.Num):
                const_index = expr.slice.value.n
                if isinstance(expr.value, ast.Name) and \
                   expr.value.id in self.arrays:
                    self.change = True
                    return self.arrays[expr.value.id].elts[const_index]
                elif isinstance(expr.value, ast.List):
                    self.change = True
                    return expr.value.elts[const_index]
            elif isinstance(expr.slice.value, ast.Str):
                const_str = expr.slice.value.s
                try:
                    if isinstance(expr.value, ast.Name) and \
                       expr.value.id in self.dicts:
                        elt = self._dict_lookup(self.dicts[expr.value.id],
                                                const_str)
                        self.change = True
                        return elt
                    elif isinstance(expr.value, ast.Dict):
                        elt = self._dict_lookup(expr.value, const_str)
                        self.change = True
                        return elt
                except KeyError:
                    pass

        self.generic_visit(expr)
        return expr



import dis
import sys
from collections import defaultdict
from tempfile import NamedTemporaryFile
from .common import try_copy

FILE_PREFIX = 'inline'


def compile_and_exec(code, globls):
    """

    """
    with NamedTemporaryFile(delete=False, prefix=FILE_PREFIX) as f:
        f.write(code.encode('utf-8'))
        f.flush()
        exec(compile(code, f.name, 'exec'), globls, globls)


# https://blog.hakril.net/articles/2-understanding-python-execution-tracer.html
class FrameAnalyzer:
    """
    Disassembles a runtime stack frame.
    """
    def __init__(self, frame):
        self.frame = frame
        self.code = frame.f_code
        self.bytecode = dis.Bytecode(self.code)
        self.cur_i = self.frame.f_lasti

    def current_instr(self):
        for i in self.bytecode:
            if i.offset == self.cur_i:
                return i

    def next_instr(self):
        for i in self.bytecode:
            if i.offset > self.cur_i:
                return i


class ValueUnknown:
    pass


class Tracer:
    """
    Executes a program and collects information about loads, stores, and executed lines.
    """
    def __init__(self,
                 prog,
                 globls=None,
                 trace_lines=False,
                 trace_opcodes=False,
                 debug=False):
        self.prog = prog
        self.reads = defaultdict(list)
        self.stores = defaultdict(list)
        self.set_count = defaultdict(int)
        self.execed_lines = defaultdict(int)
        self.trace_opcodes = trace_opcodes
        self.trace_lines = trace_lines
        self._last_store = None
        self.debug = debug
        self.globls = globls.copy() #{k: try_copy(v) for k, v in globls.items()}

    def get_value(self, frame, name):
        # Lookup value in relevant store
        if name in frame.f_locals:
            value = frame.f_locals[name]
        elif name in frame.f_globals:
            value = frame.f_globals[name]
        else:
            value = ValueUnknown()

        # Attempt to copy the value if possible
        return try_copy(value)

    def _trace_fn(self, frame, event, arg):
        if self.trace_opcodes and \
           frame.f_code.co_filename == self._fname:
            frame.f_trace_opcodes = True
        else:
            frame.f_trace_opcodes = False

        if self.trace_lines and \
           frame.f_code.co_filename == self._fname:
            frame.f_trace_lines = True
        else:
            frame.f_trace_lines = False

        if event == 'opcode':
            # The effect of a store appearing in f_locals/globals seems
            # to only happen after f_lasti advances beyond the STORE_NAME
            # so we have to cache the effect and check for it later
            if self._last_store is not None:
                name, lineno = self._last_store
                self.stores[name].append((self.get_value(frame, name), lineno))
                self._last_store = None

            instr = FrameAnalyzer(frame).current_instr()
            name = instr.argval

            if self.debug:
                print(instr.opname, str(instr.argval)[:30])

            if instr.opname in ['STORE_NAME', 'STORE_FAST', 'STORE_GLOBAL']:
                self._last_store = (name, frame.f_lineno)
                self.set_count[name] += 1
            elif instr.opname in ['LOAD_NAME', 'LOAD_FAST', 'LOAD_GLOBAL']:
                self.reads[name].append((self.get_value(frame,
                                                        name), frame.f_lineno))

        elif event == 'line' and frame.f_code.co_filename == self._fname:
            self.execed_lines[frame.f_lineno] += 1

        return self._trace_fn

    def trace(self):
        """
        Execute the provided program.

        In order for introspection tools like inspect.getsource to work on top-level
        objects, we have to actually write the code to a file and compile the code with
        the filename.
        """
        with NamedTemporaryFile(delete=False, prefix=FILE_PREFIX) as f:
            f.write(self.prog.encode('utf-8'))
            f.flush()
            self._fname = f.name

            should_trace = self.trace_lines or self.trace_opcodes
            try:
                prog_bytecode = compile(self.prog, f.name, 'exec')

                if should_trace:
                    sys.settrace(self._trace_fn)

                # Can't seem to access local variables in a list comprehension?
                # import x; [x.foo() for _ in range(10)]
                # https://github.com/inducer/pudb/issues/103
                # For now, just only use globals
                exec(prog_bytecode, self.globls, self.globls)

                if should_trace:
                    sys.settrace(None)
            except Exception:
                print(self.prog)
                raise

        return self



import ast
import inspect
from collections import defaultdict
from iterextras import unzip
import textwrap
import os
import typing
from pprint import pprint
from astpretty import pprint as pprintast
import itertools
import copy
import re
from timeit import default_timer as now
from functools import wraps, partial

from .common import *
from .visitors import *
from .tracer import *
from .passes import *


class InlineTarget:
    """
    A representation of a kind of Python object to be inlined.
    """
    def __init__(self, target):
        self.target = target

    def should_inline(self, obj):
        raise NotImplementedError


class ModuleTarget(InlineTarget):
    """
    Inline all objects defined within a module.

    e.g. if target = a.b, then objs defined in a.b or a.b.c will be inlined
    """
    def should_inline(self, obj):
        # Check if object is defined in the same module or a submodule
        # of the target.
        module = inspect.getmodule(obj)
        module_parts = module.__name__.split('.')
        target_parts = self.target.split('.')
        return module_parts[:len(target_parts)] == target_parts


class FunctionTarget(InlineTarget):
    """
    Inline exactly this function
    """
    def should_inline(self, obj):
        if inspect.ismethod(obj):
            # Get the class from the instance bound to the method
            cls = obj.__self__.__class__

            # Check if the class has the target method, and the runtime
            # object is the same
            cls_has_method = hasattr(cls, self.target.__name__)
            if cls_has_method:
                method_same_as_target = getattr(
                    cls, self.target.__name__) == self.target
                return method_same_as_target

            return False

        elif inspect.isfunction(obj):
            # If it's a normal function, then directly compare for function
            # equality
            return obj == self.target


class ClassTarget(InlineTarget):
    """
    Inline this class and all of its methods
    """
    def should_inline(self, obj):
        # e.g. target()
        try:
            constructor = self.target == obj or issubclass(obj, self.target)
        except Exception:
            constructor = False

        # e.g. f = target(); f.foo()
        bound_method = inspect.ismethod(obj) and isinstance(
            obj.__self__, self.target)

        # e.g. f = target(); target.foo(f)
        # https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3
        if inspect.isfunction(obj):
            qname = obj.__qualname__.split('.')
            unbound_method = len(
                qname) > 1 and qname[-2] == self.target.__name__
        else:
            unbound_method = False

        # e.g. f = target(); f()
        dunder_call = isinstance(obj, self.target)

        return constructor or bound_method or unbound_method or dunder_call


class Inliner:
    def __init__(self, source, targets, globls=None, unwrap_function=False):
        if not isinstance(source, str):
            if globls is None and hasattr(source, '__globals__'):
                globls = {**source.__globals__, **get_function_locals(source)}

            source = inspect.getsource(source)

        mod = ast.parse(textwrap.dedent(source))
        if len(mod.body) == 1 and isinstance(mod.body[0], ast.FunctionDef) \
           and unwrap_function:
            body = mod.body[0].body
        else:
            body = mod.body
        self.module = ast.Module(body=body)

        self.globls = globls.copy() if globls is not None else {}

        self.generated_vars = defaultdict(int)
        self.targets = []
        for target in targets:
            self.add_target(target)

        self.history = [(copy.deepcopy(self.module), None)]

        finder = FindAssignments()
        finder.visit(self.module)
        self.toplevel_vars = finder.names

        def make_pass_name(name):
            # Split "TheFooPass" into ["The", "Foo", "Pass"]
            parts = re.findall('.[^A-Z]*', name)

            # Drop "Pass"
            parts = parts[:-1]

            # Make "the_foo"
            return '_'.join([s.lower() for s in parts])

        for pass_ in PASSES:
            name = make_pass_name(pass_.__name__)
            fn = partial(self.run_pass, pass_)
            fn.__name__ = name
            setattr(self, name, fn)

        self.profiling_data = defaultdict(list)
        self._tracer_cache = None

    def _make_target(self, target):
        if isinstance(target, str):
            return ModuleTarget(target)
        elif inspect.isfunction(target):
            return FunctionTarget(target)
        elif inspect.isclass(target):
            return ClassTarget(target)
        elif isinstance(target, InlineTarget):
            return target
        else:
            raise Exception(
                "Can't make inline target from object: {}".format(target))

    def add_target(self, target):
        target = self._make_target(target)
        self.targets.append(target)

    def fresh(self, prefix='var'):
        """
        Creates a new variable semi-guaranteed to not exist in the program.
        """
        self.generated_vars[prefix] += 1
        count = self.generated_vars[prefix]
        if count == 1:
            return f'{prefix}'
        else:
            return f'{prefix}_{count}'

    def is_source_obj(self, obj):
        """
        Checks if runtime object was defined in the inliner source.

        Requires that the executed code was run through the tracer.
        """
        try:
            if os.path.basename(inspect.getfile(obj)).startswith(FILE_PREFIX):
                return True
        except TypeError:
            pass

        return False

    def should_inline(self, obj):
        """
        Checks whether a runtime object is something to be inlined.
        """
        module = inspect.getmodule(obj)

        # Unconditionally inline objects defined in the source
        if self.is_source_obj(obj):
            return True

        # Don't inline builtins (they don't have source)
        if inspect.isbuiltin(obj):
            return False

        # Don't inline objects without a module
        if module is None:
            return False

        for target in self.targets:
            if target.should_inline(obj):
                return True

        return False

    def run_pass(self, Pass, **kwargs):
        start = now()
        try:
            pass_ = Pass(self)
            change = pass_.run(**kwargs)
        except Exception:
            self.module = copy.deepcopy(self.history[-1][0])
            raise
        end = now() - start

        if Pass.tracer_args is not None and not change:
            self._tracer_cache = (Pass.tracer_args, pass_.tracer)
        else:
            self._tracer_cache = None

        self.profiling_data[Pass.__name__].append(end)
        self.history.append((copy.deepcopy(self.module), Pass.__name__))

        return change

    def undo(self):
        self.history.pop()
        self.module = copy.deepcopy(self.history[-1][0])

    def make_program(self, comments=False):
        return a2s(self.module, comments=comments).rstrip()

    def fixpoint(self, f):
        change = False
        while f():
            change = True
            pass
        return change

    def modules(self):
        tracer = self.execute()
        collector = CollectModules(tracer.globls)
        collector.visit(self.module)
        return collector.modules

    def autoschedule(self):
        while True:
            if not self.inline():
                break
            self.fixpoint(self.deadcode)

        while True:
            any_pass = any([
                self.unread_vars(),
                self.lifetimes(),
                self.expand_self(),
                self.copy_propagation(),
                self.simplify_varargs(),
                self.expand_tuples()
            ])
            if not any_pass:
                break

        self.clean_imports()
        self.remove_suffixes()

    def execute(self):
        return Tracer(self.make_program(), self.globls).trace()



from .array_index import ArrayIndexPass
from .clean_imports import CleanImportsPass
from .copy_propagation import CopyPropagationPass
from .deadcode import DeadcodePass
from .expand_self import ExpandSelfPass
from .expand_tuples import ExpandTuplesPass
from .inline import InlinePass
from .lifetimes import LifetimesPass
from .partial_eval import PartialEvalPass
from .remove_suffixes import RemoveSuffixesPass
from .simplify_varargs import SimplifyVarargsPass
from .unread_vars import UnreadVarsPass

PASSES = [
    ArrayIndexPass,
    CleanImportsPass,
    CopyPropagationPass,
    DeadcodePass,
    ExpandSelfPass,
    ExpandTuplesPass,
    InlinePass,
    LifetimesPass,
    PartialEvalPass,
    RemoveSuffixesPass,
    SimplifyVarargsPass,
    UnreadVarsPass
] # yapf: disable



import ast
from collections import defaultdict
import sys

from .base_pass import BasePass
from ..visitors import CollectLineNumbers
from ..common import is_effect_free


class LifetimesPass(BasePass):
    """
    Lifetime/liveness analysis to eliminate unused variables.

    Analyzes whether a variable is read before being re-assigned, and eliminates
    the assignment if unused.

    Example:
      x = 1
      x = 2
      assert x == 2

      >> becomes >>

      x = 2
      assert x == 2
    """

    tracer_args = {'trace_opcodes': True}

    def __init__(self, inliner):
        super().__init__(inliner)
        self._find_unused_stores()

    def _find_unused_stores(self):
        self.unused_stores = defaultdict(list)
        for k, stores in self.tracer.stores.items():
            reads = self.tracer.reads[k]

            for i, (_, store_line) in enumerate(stores):
                next_possible_stores = [
                    line for _, line in stores[i + 1:] if line > store_line
                ]
                next_store_line = min(next_possible_stores, default=sys.maxsize)

                unused = True
                for (_, read_line) in reads:
                    if store_line < read_line and read_line <= next_store_line:
                        unused = False

                if unused:
                    self.unused_stores[k].append(store_line)

    def visit_ImportFrom(self, stmt):
        aliases = [
            alias for alias in stmt.names
            if len(self.tracer.reads[alias.name if alias.
                                     asname is None else alias.asname]) > 0
        ]

        if len(aliases) != len(stmt.names):
            self.change = True
            return None

        if len(aliases) > 0:
            stmt.names = aliases
            return stmt

    def visit_FunctionDef(self, stmt):
        func_unread = len(self.tracer.reads[stmt.name]) == 0
        no_decorators = len(stmt.decorator_list) == 0
        if func_unread and no_decorators:
            self.change = True
            return None
        return stmt

    def visit_Assign(self, stmt):
        if len(stmt.targets) == 1 and \
           isinstance(stmt.targets[0], ast.Name):

            name = stmt.targets[0].id
            unused_lines = self.unused_stores[name]

            collect_lineno = CollectLineNumbers()
            collect_lineno.visit(stmt)

            if len(set(unused_lines) & collect_lineno.linenos) > 0 and \
               is_effect_free(stmt.value):
                self.change = True
                return None

        return stmt



import ast

from ..common import compare_ast
from .base_pass import BasePass


class CleanImportsPass(BasePass):
    """
    Puts all imports at the top of the module and de-duplicates them.

    Example:
      x = 1
      import foo
      import numpy
      y = foo.bar() + x

      >> becomes >>

      import foo
      x = 1
      y = foo.bar() + x
    """
    def __init__(self, inliner):
        super().__init__(inliner)
        self.imports = []

    def generic_visit(self, node):
        # Collect all imports and remove them from the module
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            self.imports.append(node)
            return None

        return super().generic_visit(node)

    def after_visit(self, mod):
        # Use exact AST equality to deduplicate imports
        imports_dedup = [
            imprt for i, imprt in enumerate(self.imports) if not any(
                [compare_ast(imprt, imprt2) for imprt2 in self.imports[i + 1:]])
        ]

        # Add imports back to the top of the module
        mod.body = imports_dedup + mod.body



import ast

from .base_pass import BasePass
from ..common import a2s, obj_to_ast, compare_ast


class ShouldEval(ast.NodeVisitor):
    def __init__(self):
        self.should = True

    def generic_visit(self, node):

        whitelist = (ast.Num, ast.Str, ast.List, ast.Tuple, ast.Dict,
                     ast.NameConstant, ast.Name, ast.Load, ast.UnaryOp,
                     ast.BinOp, ast.BoolOp, ast.Compare, ast.Attribute,
                     ast.Subscript, ast.Index, ast.Slice, ast.ExtSlice, ast.Eq,
                     ast.NotEq, ast.Is, ast.In, ast.Add, ast.Sub, ast.Mult,
                     ast.Div, ast.And, ast.Or)

        if not isinstance(node, whitelist):
            self.should = False

        return super().generic_visit(node)


class PartialEvalPass(BasePass):
    def generic_visit(self, node):
        should_eval = ShouldEval()
        should_eval.visit(node)

        if should_eval.should:
            try:
                src = a2s(node)
                ret = eval(src, {}, {})
                retnode = obj_to_ast(ret)
                return retnode
            except Exception:
                pass

        super().generic_visit(node)
        return node



import ast
from .base_pass import BasePass
from ..common import robust_eq, is_effect_free, tree_size
from ..visitors import Replace

MAX_TREESIZE = 10  # TODO: good value for this?


class CopyPropagationPass(BasePass):
    """
    Copies values read once into their point of use.

    Copy propagation has two components: CAN a value be propagated, and
    SHOULD it be.

    * CAN: copy propagation is challenging in the presence of an effect. Most
           definitions cannot be trivially inlined. For now, I define a
           variable assignment can be safely propagated if any are true:
      - the variable is not modified during its execution, i.e.
        its value on the first read is equal to the last read
      - it's read exactly once (or generally N times, under a loop with
        N iterations)
      - the assignment is a direct alias, e.g. x = y

    * SHOULD: even if a propagation is safe, arbitrarily inlining values
              can make the code difficult to read. Variables are useful
              tools to break up long segments of code. So I use a few
              heuristics to determine whether a syntax object can be inlined.
      - The size of its syntax tree is less than MAX_TREESIZE nodes
      - The variable must be one that was generated during inlining, not from
        the original script

    Example:
      x = {'a': 1}
      y = x['a']
      assert y == 1

      >> becomes >>

      assert {'a': 1}['a'] == 1

    Example:
      sum = 0
      for i in range(10):
        x = i * 10
        sum += x

      >> becomes >>

      sum = 0
      for i in range(10):
        sum += i * 10
    """

    tracer_args = {'trace_opcodes': True, 'trace_lines': True}

    def __init__(self, inliner):
        super().__init__(inliner)
        self.assignments = []
        self.baseline_execs = 1

    def visit_With(self, stmt):
        # For now, don't do any propagation on assignments within with blocks
        # to avoid moving statements outside the block
        return stmt

    def visit_For(self, loop):
        # Track the current number of loop iterations as we descend the AST
        loop_iters = self.tracer.execed_lines[loop.lineno] - 1
        if loop_iters > 0:
            self.baseline_execs *= loop_iters
            self.generic_visit(loop)
            self.baseline_execs /= loop_iters
        return loop

    def visit_Assign(self, stmt):
        if len(stmt.targets) == 1 and \
           isinstance(stmt.targets[0], ast.Name):
            k = stmt.targets[0].id

            # Attempt to determine whether the variable's value changed
            # during tracing. robust_eq is... marginally robust.
            try:
                val_eq = len(self.tracer.reads[k]) == 0 or \
                   bool(robust_eq(self.tracer.reads[k][0][0],
                                  self.tracer.reads[k][-1][0]))
            except Exception as e:
                # TODO: do some better kind of logging here
                print('WARNING: could not compare for equality on key `{}`:'.
                      format(k))
                print('Inital', str(self.tracer.reads[k][0][0])[:100])
                print('Final', str(self.tracer.reads[k][-1][0])[:100])
                print(e)
                print('=' * 30)

                val_eq = False

            can_propagate = (val_eq or
                len(self.tracer.reads[k]) == self.baseline_execs or
                isinstance(stmt.value, ast.Name)) and \
                is_effect_free(stmt.value)

            should_propagate = k not in self.inliner.toplevel_vars and \
                tree_size(stmt.value) <= MAX_TREESIZE

            if self.tracer.set_count[k] == self.baseline_execs and \
               can_propagate and should_propagate:
                self.assignments.append((k, stmt.value))
                return None

        return stmt

    def after_visit(self, mod):
        # Once we have collected the copyable assignments, go through and
        # replace every usage of them
        for i, (name, value) in enumerate(self.assignments):
            replacer = Replace(name, value)
            replacer.visit(mod)

            # Have to update not just the main AST, but also any copyable
            # assignments that might reference copies. For example:
            #
            # x = 1
            # y = x
            # z = y
            #
            # After copying x, self.assignments will still have y = x, so
            # a naive copy of y = x into z will then produce the program z = x
            # with no definition of x.
            for j, (name2, value2) in enumerate(self.assignments[i + 1:]):
                self.assignments[i + 1 + j] = (name2, replacer.visit(value2))

        self.change = len(self.assignments) > 0



import ast
import inspect

from .base_pass import BasePass
from ..common import a2s, parse_expr, make_name, parse_stmt
from ..transforms import ContextualTransforms


class FindCall(ast.NodeTransformer):
    def __init__(self, inliner, globls):
        self.call_obj = None
        self.call_expr = None
        self.inliner = inliner
        self.globls = globls
        self.ret_var = None

    def visit_Call(self, call_expr):
        # for example, self.foo(1)
        try:
            # get the runtime object for function self.foo
            call_obj = eval(a2s(call_expr.func), self.globls, self.globls)
        except Exception:
            return call_expr

        # If a function was generated by a higher-order function, we can't
        # directly inline it, must be inlined through the generator. This
        # is detected by checking if the function object has a closure.
        if not self.inliner.is_source_obj(call_obj) and \
           hasattr(call_obj, '__closure__') and \
           call_obj.__closure__ is not None:

            fdef = parse_stmt(inspect.getsource(call_obj))
            if len(fdef.decorator_list) == 0:
                print(
                    'found closure', a2s(call_expr),
                    list(
                        zip(call_obj.__code__.co_freevars,
                            call_obj.__closure__)))
                self.generic_visit(call_expr)
                return call_expr

        if not isinstance(call_expr.func, ast.Call) and \
           self.inliner.should_inline(call_obj):
            self.call_expr = call_expr
            self.call_obj = call_obj

            # Heuristically make a good name for the drop-in variable
            func = call_expr.func
            if isinstance(func, ast.Name):
                func_name = func.id
            elif isinstance(func, ast.Attribute):
                func_name = func.attr
            else:
                func_name = 'func'

            self.ret_var = self.inliner.fresh(f'{func_name}_ret')
            return make_name(self.ret_var)

        self.generic_visit(call_expr)
        return call_expr

    def generic_visit(self, node):
        if self.call_expr is not None:
            return node
        return super().generic_visit(node)

    def _is_property(self, node):
        assert isinstance(node, ast.Attribute)

        # for example, foo.x where the class of foo has @property def x()
        try:
            # get the runtime object for foo
            prop_obj = eval(a2s(node.value), self.globls, self.globls)
        except Exception:
            # if we can't find it, ignore
            return None

        # if foo should be inlined, and it is an instance of a class,
        # and the class has the attribute, and the attribute is a property
        if self.inliner.should_inline(prop_obj) and \
           hasattr(prop_obj, '__class__') and \
           hasattr(prop_obj.__class__, node.attr):
            prop = getattr(prop_obj.__class__, node.attr)
            if isinstance(prop, property):
                return (prop, prop_obj)

        return None

    def visit_Assign(self, assgn):
        if len(assgn.targets) == 1:
            target = assgn.targets[0]
            if isinstance(target, ast.Attribute):
                ret = self._is_property(target)
                if ret is not None:
                    (prop, prop_obj) = ret
                    self.call_obj = prop.fset
                    self.call_expr = parse_expr("{}.{}_setter({}, {})".format(
                        prop_obj.__class__.__name__, target.attr,
                        a2s(target.value), a2s(assgn.value)))
                    self.ret_var = self.inliner.fresh(target.attr)
                    return None

        self.generic_visit(assgn)
        return assgn

    # Classes using @property have accessors that are actually calling
    # functions. This visitor looks for uses of @property.
    def visit_Attribute(self, attr):
        ret = self._is_property(attr)
        if ret is not None:
            (prop, prop_obj) = ret
            # foo.x is same as Foo.x.fget(foo), so we treat the property
            # as a function call so we can pass it to the function inliner
            self.call_obj = prop.fget
            self.call_expr = parse_expr("{}.{}_getter({})".format(
                prop_obj.__class__.__name__, attr.attr, a2s(attr.value)))

            # TODO: need to generate an import for
            # prop_obj.__class__.__name__

            self.ret_var = self.inliner.fresh(attr.attr)
            return make_name(self.ret_var)

        self.generic_visit(attr)
        return attr


class FindComprehension(ast.NodeTransformer):
    def __init__(self, find_call, ret_var):
        self.find_call = find_call
        self.ret_var = ret_var
        self.comp = None

    def visit_ListComp(self, comp):
        self.find_call.visit(comp)
        if self.find_call.call_expr is not None:
            self.comp = comp
            return make_name(self.ret_var)
        else:
            return comp


class FindIfExp(ast.NodeTransformer):
    def __init__(self, inliner):
        self.inliner = inliner
        self.ifexp = None
        self.ret_var = None
        self.assign_name = None

    def visit_Assign(self, stmt):
        if len(stmt.targets) == 1 and \
           isinstance(stmt.targets[0], ast.Name):
            self.assign_name = stmt.targets[0].id
            self.generic_visit(stmt)
            self.assign_name = None
        else:
            self.generic_visit(stmt)
        return stmt

    def visit_IfExp(self, ifexp):
        self.ifexp = ifexp
        name = self.assign_name + '_ifexp' if self.assign_name is not None else 'ifexp'
        self.ret_var = self.inliner.fresh(name)
        return make_name(self.ret_var)


class FindAssignment(ast.NodeVisitor):
    def __init__(self, var):
        self.assgn = None
        self.var = var

    def visit_Assign(self, stmt):
        if len(stmt.targets) == 1 and \
           isinstance(stmt.targets[0], ast.Name) and \
           stmt.targets[0].id == self.var:
            self.assgn = stmt.value


class InlinePass(BasePass):
    """
    Primary inlining functionality. Looks for inline-able syntax objects
    and inlines them. Currently inlines:
    - Functions
    - Objects (constructors and methods)
    - Generators
    - With blocks
    - List comprehensions
    - If-expressions

    Basic strategy is to start at a statement, e.g. an assignment, then
    search the assignment for inlineable objects. If one is found, it is
    replaced by a fresh variable, and the logic to assign that variable
    is placed above the statement.

    Example:
      def foo():
        return 1
      assert foo() + 1 == 2

      >> becomes (roughly) >>

      foo_ret = 1
      assert foo_ret + 1 == 2

    Most of the code in this module just finds the relevant objects. The
    code to perform the inlining lives in the ContextualTransforms class
    in transforms.py.
    """
    tracer_args = {}

    def __init__(self, inliner):
        super().__init__(inliner)
        self.fns = ContextualTransforms(self.inliner, self.globls)

    def visit_For(self, stmt):
        self.generic_visit(stmt)

        mod = ast.Module([ast.Expr(stmt.iter)])
        self.visit(mod)
        stmt.iter = mod.body[-1].value

        return mod.body[:-1] + [stmt]

    def visit_With(self, stmt):
        self.generic_visit(stmt)

        context = eval(a2s(stmt.items[0].context_expr), self.globls,
                       self.globls)
        if self.inliner.should_inline(context):
            self.change = True
            return self.fns.expand_with(stmt)
        else:
            return [stmt]

    def visit_FunctionDef(self, fdef):
        if len(fdef.decorator_list) > 0:
            name = make_name(fdef.name)
            dec_assgn = ast.Assign(targets=[name],
                                   value=ast.Call(func=fdef.decorator_list[0],
                                                  args=[name],
                                                  keywords={}))
            fdef.decorator_list = []
            return [fdef, dec_assgn]
        return fdef

    def _visit_import(self, imprt):
        if len(imprt.names) == 1:
            alias = imprt.names[0]
            source_name = alias.name
            imported_name = alias.asname or alias.name
            try:
                obj = eval(imported_name, self.globls, self.globls)
            except Exception:
                return imprt

            if self.inliner.should_inline(obj) and \
               not inspect.ismodule(obj) and \
               not inspect.isclass(obj):

                if isinstance(imprt, ast.ImportFrom):
                    assert imprt.level == 0
                    exec(f'import {imprt.module}', self.globls, self.globls)
                    mod_obj = eval(imprt.module, self.globls, self.globls)
                else:
                    assert False, "TODO"

                mod_ast = ast.parse(open(inspect.getsourcefile(mod_obj)).read())
                finder = FindAssignment(source_name)
                finder.visit(mod_ast)

                # TODO: generate for finder.assgn
                if finder.assgn is not None:
                    return ast.Assign(targets=[make_name(imported_name)],
                                      value=finder.assgn)

        return imprt

    def generic_visit(self, node):
        if isinstance(node, (ast.Assert, ast.Assign, ast.Expr)):
            return self._inline(node)

        if isinstance(node, (ast.Import, ast.ImportFrom)):
            return self._visit_import(node)

        return super().generic_visit(node)

    def _inline(self, stmt):
        ifexp_finder = FindIfExp(self.inliner)
        ifexp_finder.visit(stmt)
        if ifexp_finder.ifexp is not None:
            self.change = True
            return self.fns.expand_ifexp(ifexp_finder.ifexp,
                                         ifexp_finder.ret_var) + [stmt]

        comp_ret_var = self.inliner.fresh('comp')
        comp_call_finder = FindCall(self.inliner, self.globls)
        comp_finder = FindComprehension(comp_call_finder, comp_ret_var)
        comp_finder.visit(stmt)
        if comp_finder.comp is not None:
            self.change = True
            return self.fns.expand_comprehension(comp_finder.comp, comp_ret_var,
                                                 comp_call_finder) + [stmt]

        call_finder = FindCall(self.inliner, self.globls)
        stmt = call_finder.visit(stmt)

        new_stmts = []
        if call_finder.call_expr is not None:
            ret_var = call_finder.ret_var
            call_expr = call_finder.call_expr
            call_obj = call_finder.call_obj

            if inspect.ismethod(call_obj):
                imprt = self.fns.expand_method(call_obj, call_expr, ret_var)
                if imprt is not None:
                    new_stmts.insert(0, imprt)
                new_stmts.append(
                    ast.Assign(targets=[make_name(ret_var)], value=call_expr))

            elif inspect.isgeneratorfunction(call_obj):
                new_stmts.extend(
                    self.fns.inline_generator_function(call_obj, call_expr,
                                                       ret_var))

            elif inspect.isfunction(call_obj):
                new_stmts.extend(
                    self.fns.inline_function(call_obj,
                                             call_expr,
                                             ret_var,
                                             debug=self.args['debug']))

            elif inspect.isclass(call_obj):
                new_stmts.extend(
                    self.fns.inline_constructor(call_obj, call_expr, ret_var))

            elif hasattr(call_obj, '__call__'):
                self.fns.expand_callable(call_expr)
                new_stmts.append(
                    ast.Assign(targets=[make_name(ret_var)], value=call_expr))

            else:
                print(call_obj, type(call_obj), a2s(call_expr).strip())
                raise NotYetImplemented

            self.change = True

        new_stmts.append(stmt)
        return new_stmts



import ast
from .base_pass import BasePass


class ExpandTuplesPass(BasePass):
    """
    Expand tuple assignments to multiple variable assignments.

    Useful for making tuple assignments optimizable by other passes like copy
    propagation.

    Example:
      x, y = (1, 2)

      >> becomes >>

      x = 1
      y = 2
    """
    def visit_Assign(self, stmt):
        if isinstance(stmt.value, ast.Tuple):
            # x, y = tuple
            if len(stmt.targets) > 1:
                targets = stmt.targets
            # (x, y) = tuple
            elif isinstance(stmt.targets[0], ast.Tuple):
                targets = stmt.targets[0].elts
            else:
                return stmt

            self.change = True
            return [
                ast.Assign(targets=[name], value=elt)
                for name, elt in zip(targets, stmt.value.elts)
            ]

        return stmt



import ast
from .base_pass import BasePass


# TODO: is this redundant with lifetimes?
class UnreadVarsPass(BasePass):
    tracer_args = {'trace_opcodes': True}

    def visit_Assign(self, stmt):
        if len(stmt.targets) == 1 and \
           isinstance(stmt.targets[0], ast.Name):
            name = stmt.targets[0].id
            if len(self.tracer.reads[name]) == 0:
                self.change = True
                return None

        return stmt

    def visit_ImportFrom(self, stmt):
        aliases = [
            alias for alias in stmt.names
            if len(self.tracer.reads[alias.name if alias.
                                     asname is None else alias.asname]) > 0
        ]

        if len(aliases) != len(stmt.names):
            self.change = True
            return None

        if len(aliases) > 0:
            stmt.names = aliases
            return stmt

    def visit_FunctionDef(self, stmt):
        if len(self.tracer.reads[stmt.name]) == 0:
            self.change = True
            return None

        return stmt



import inspect
import ast
import textwrap
import itertools
from iterextras import unzip
import typing

from .visitors import RemoveFunctoolsWraps, ReplaceYield, UsedGlobals, \
    ReplaceSuper, ReplaceSelf, Rename, FindAssignments, ReplaceReturn, \
    collect_imports
from .common import a2s, parse_stmt, parse_expr, make_name, obj_to_ast, SEP, COMMENT_MARKER


class ContextualTransforms:
    """
    Primary code for performing inlining.

    Inlining code often needs to make reference to either the Inliner object
    or to the global variables generated during the last trace, so we wrap all
    the functions in a ContextualTransforms class that provides access to this
    state.
    """
    def __init__(self, inliner, globls):
        self.inliner = inliner
        self.globls = globls

    def inline_constructor(self, call_obj, call_expr, ret_var):
        """
        Inlines a class constructor.

        Construction has two parts: creating the object with __new__, and
        initializing it with __init__. We insert a __new__ call and then
        inline the __init__ function.

        Example:
          class Foo:
            def __init__(self):
              self.x = 1
          f = Foo()

          >> becomes >>

          f = Foo.__new__(Foo)
          self = f
          self.x = 1
        """
        cls = call_obj.__name__

        # Add an import for the class
        imprt = ast.ImportFrom(module=inspect.getmodule(call_obj).__name__,
                               level=0,
                               names=[ast.alias(name=cls, asname=None)])

        # Create a raw object using __new__
        make_obj = ast.Assign(targets=[make_name(ret_var)],
                              value=parse_expr(f'{cls}.__new__({cls})'))

        # Add the object as an explicit argument to the __init__ function
        call_expr.args.insert(0, make_name(ret_var))

        if call_obj.__init__ is not object.__init__:
            # Inline the __init__ function
            init_inline = self.inline_function(call_obj.__init__,
                                               call_expr,
                                               ret_var,
                                               cls=call_obj)
        else:
            init_inline = []

        return [imprt, make_obj] + init_inline

    def expand_method(self, call_obj, call_expr, ret_var):
        """
        Replace bound methods with unbound functions.

        Example:
          f = Foo()
          assert f.bar(0) == 1

          >> becomes >>

          f = Foo()
          assert Foo.bar(f, 0) == 1
        """

        call_obj_ast = call_expr.func.value

        # HACK: assume all methods are called syntactically as obj.method()
        # as opposed to x = obj.method; x()
        assert isinstance(call_expr.func, ast.Attribute)

        # Get the object bound to the method
        bound_obj = call_obj.__self__

        # If the method is a classmethod, the method is bound to the class
        if inspect.isclass(bound_obj):
            # The unbound function underlying a classmethod can be accessed
            # through Foo.x.__func__
            clsmethod = ast.Attribute(value=make_name(bound_obj.__name__),
                                      attr=call_expr.func.attr)
            call_expr.func = ast.Attribute(value=clsmethod, attr='__func__')
        else:
            # Rewrite foo.x to Foo.x
            cls = bound_obj.__class__
            call_expr.func = ast.Attribute(value=make_name(cls.__name__),
                                           attr=call_expr.func.attr)

        # Add the object as explicit self parameter
        call_expr.args.insert(0, call_obj_ast)

        # Generate any imports needed
        file_imports = collect_imports(call_obj)
        return self.generate_imports(cls.__name__,
                                     cls,
                                     call_obj=call_obj,
                                     file_imports=file_imports)

    def expand_callable(self, call_expr):
        """
        Expands uses of __call__.

        Example:
          f = Foo()
          assert f() == 1

          >> becomes >>

          f = Foo()
          assert Foo.__call__(f) == 1
        """
        call_expr.func = ast.Attribute(value=call_expr.func, attr='__call__')

    def inline_generator_function(self, call_obj, call_expr, ret_var):
        """
        Inlines generator functions (those using yield).

        There is no easy way to proxy generator semantics/control flow
        without generators, unlike early returns. The simple strategy is to
        eagerly materialize the generator into a list. However, this is both
        inefficient and does not always preserve semantics, e.g. see
        requests.ipynb.

        Example:
          def foo():
            for i in range(10):
              yield i
          for i in foo():
             print(i)

          >> becomes >>
          l = []
          for i in range(10):
            l.append(i)
          for i in l:
            print(i)
        """
        f_ast = parse_stmt(textwrap.dedent(inspect.getsource(call_obj)))

        # Initialize the list
        new_stmts = [parse_stmt(f'{ret_var} = []')]

        # Replace all yield statements by appending to the list
        ReplaceYield(ret_var).visit(f_ast)

        # Then inline the function as normal
        new_stmts.extend(
            self.inline_function(call_obj, call_expr, ret_var, f_ast=f_ast))
        return new_stmts

    def _expand_decorators(self, new_stmts, f_ast, call_expr, call_obj,
                           ret_var):
        """
        Expand decorator calls to an inlined function.

        Example:
          @foo
          def bar(x):
            return x + 1
          assert bar(1) == 2

          >> becomes >>

          def bar(x):
            return x + 1
          assert foo(bar(1)) == 2
        """
        decorators = f_ast.decorator_list

        used_globals = UsedGlobals(call_obj.__globals__)
        used_globals.visit(f_ast)
        used = used_globals.used
        file_imports = collect_imports(call_obj)
        for name, globl in used_globals.used.items():
            imprt = self.generate_imports(name,
                                          globl,
                                          call_obj=call_obj,
                                          file_imports=file_imports)
            if imprt is not None:
                new_stmts.insert(0, imprt)

        f_ast.decorator_list = []
        new_stmts.append(f_ast)

        # TODO: what if decorator has arguments?
        call_expr.func = ast.Call(func=decorators[0],
                                  args=[make_name(f_ast.name)],
                                  keywords=[])
        new_stmts.append(
            ast.Assign(targets=[make_name(ret_var)], value=call_expr))

    def _replace_self_super(self, f_ast, cls, call_expr):
        # If we don't know what the class is, e.g. in Foo.method(foo), then
        # eval the LHS of the attribute, e.g. Foo here
        if cls is None:
            if isinstance(call_expr.func, ast.Attribute):
                cls = eval(a2s(call_expr.func.value), self.globls, self.globls)
            else:
                cls = eval(a2s(call_expr.func), self.globls,
                           self.globls).__class__

        # HACK: we're assuming super() always refers to the first base class,
        # but it actually depends on the specific method being called and the MRO.
        # THIS IS UNSOUND with multiple inheritance (and potentially even with
        # basic subtype polymorphism?)
        ReplaceSuper(cls.__bases__[0]).visit(f_ast)

        # Replace references to `self.method()` with `Foo.method(self)`, same as
        # expand_constructor()
        ReplaceSelf(cls, self.globls).visit(f_ast)

    def _bind_arguments(self, f_ast, call_expr, new_stmts):
        args_def = f_ast.args

        # Scope a variable name as unique to the function, and update any references
        # to it in the function
        def unique_and_rename(name):
            unique_name = f'{name}{SEP}{f_ast.name}'
            renamer = Rename(name, unique_name)
            for stmt in f_ast.body:
                renamer.visit(stmt)
            return unique_name

        args = call_expr.args[:]

        # Rename all variables declared in the function that aren't arguments
        assgn_finder = FindAssignments()
        assgn_finder.visit(f_ast)
        arg_names = set([arg.arg for arg in args_def.args])
        for name in assgn_finder.names:
            if name not in arg_names:
                unique_and_rename(name)

        # If function is called with f(*args)
        if len(call_expr.args) > 0 and \
           isinstance(call_expr.args[-1], ast.Starred):
            star_arg = call_expr.args.pop().value

            # Get the length of the star_arg runtime list
            star_arg_obj = eval(a2s(star_arg), self.globls, self.globls)

            # Generate an indexing expression for each element of the list
            call_star_args = [
                ast.Subscript(value=star_arg, slice=ast.Index(value=ast.Num(i)))
                for i in range(len(star_arg_obj))
            ]
        else:
            star_arg = None

        # If function is called with f(**kwargs)
        star_kwarg = [arg for arg in call_expr.keywords if arg.arg is None]
        star_kwarg = star_kwarg[0].value if len(star_kwarg) > 0 else None
        if star_kwarg is not None:
            star_kwarg_dict = eval(a2s(star_kwarg), self.globls, self.globls)
            call_star_kwarg = {
                key: ast.Subscript(value=star_kwarg,
                                   slice=ast.Index(value=ast.Str(key)))
                for key in star_kwarg_dict.keys()
            }

        # Function's anonymous arguments, e.g. f(1, 2) becomes [1, 2]
        call_anon_args = call_expr.args[:]

        # Function's keyword arguments, e.g. f(x=1, y=2) becomes {'x': 1, 'y': 2}
        call_kwargs = {
            arg.arg: arg.value
            for arg in call_expr.keywords if arg.arg is not None
        }

        # Match up defaults with variable names.
        #
        # Python convention is that if function has N arguments and K < N defaults, then
        # the defaults correspond to arguments N - K .. N.
        nodefault = len(args_def.args) - len(args_def.defaults)
        anon_defaults = {
            arg.arg: default
            for arg, default in zip(args_def.args[nodefault:],
                                    args_def.defaults)
        }

        # All keyword-only arguments must have defaults.
        #
        # kwonlyargs occur if a function definition has args AFTER a *args, e.g.
        # the var "y" in `def foo(x, *args, y=1)`
        kw_defaults = {
            arg.arg: default
            for arg, default in zip(args_def.kwonlyargs, args_def.kw_defaults)
        }

        # For each non-keyword-only argument, match it up with the corresponding
        # syntax from the call expression
        for arg in args_def.args:
            k = arg.arg

            # First, match with anonymous arguments
            if len(call_anon_args) > 0:
                v = call_anon_args.pop(0)

            # Then use *args if it exists
            elif star_arg is not None and len(call_star_args) > 0:
                v = call_star_args.pop(0)

            # Then use keyword arguments
            elif k in call_kwargs:
                v = call_kwargs.pop(k)

            # Then use **kwargs if it exists
            elif star_kwarg is not None and k in call_star_kwarg:
                v = call_star_kwarg.pop(k)

            # Otherwise use the default value
            else:
                v = anon_defaults.pop(k)

            # Add a binding from function argument to call argument
            uniq_k = unique_and_rename(k)
            stmt = ast.Assign(targets=[make_name(uniq_k)], value=v)
            new_stmts.append(stmt)

        # Perform equivalent procedure as above, but for keyword-only arguments
        for arg in args_def.kwonlyargs:
            k = arg.arg

            if k in call_kwargs:
                v = call_kwargs.pop(k)
            elif star_kwarg is not None and k in call_star_kwarg:
                v = call_star_kwarg.pop(k)
            else:
                v = kw_defaults.pop(k)

            uniq_k = unique_and_rename(k)
            stmt = ast.Assign(targets=[make_name(uniq_k)], value=v)
            new_stmts.append(stmt)

        # If function definition uses *args, then assign it to the remaining anonymous
        # arguments from the call_expr
        if args_def.vararg is not None:
            k = unique_and_rename(args_def.vararg.arg)
            v = call_anon_args[:]
            if star_arg is not None:
                v += call_star_args
            new_stmts.append(
                ast.Assign(targets=[make_name(k)], value=ast.List(elts=v)))

        # Similarly for **kwargs in the function definition
        if args_def.kwarg is not None:
            k = unique_and_rename(args_def.kwarg.arg)
            items = call_kwargs.items()
            if star_kwarg is not None:
                items = itertools.chain(items, call_star_kwarg.items())
            kwkeys, kwvalues = unzip(items)
            new_stmts.append(
                ast.Assign(targets=[make_name(k)],
                           value=ast.Dict([ast.Str(s) for s in kwkeys],
                                          kwvalues)))

    def _get_nonlocal_vars(self, f_ast, call_obj):
        used_globals = UsedGlobals(call_obj.__globals__)
        used_globals.visit(f_ast)
        used = used_globals.used

        if call_obj.__closure__ is not None and len(call_obj.__closure__) > 0:
            for var, cell in zip(call_obj.__code__.co_freevars,
                                 call_obj.__closure__):
                used[var] = cell.cell_contents

        return used

    def inline_function(self,
                        call_obj,
                        call_expr,
                        ret_var,
                        cls=None,
                        f_ast=None,
                        debug=False):
        if debug:
            print('Inlining {}'.format(a2s(call_expr)))

        # Start by adding the call expression as a comment
        new_stmts = [ast.Expr(ast.Str(COMMENT_MARKER + a2s(call_expr).strip()))]

        if f_ast is None:
            # Get the source code for the function
            try:
                f_source = inspect.getsource(call_obj)
            except TypeError:
                print('Failed to get source of {}'.format(a2s(call_expr)))
                raise

            # We have to "dedent" it if the source code is not at the top level
            # (e.g. a class method)
            f_source = textwrap.dedent(f_source)

            # Then parse the function into an AST
            f_ast = parse_stmt(f_source)

        # Give the function a fresh name so it won't conflict with other calls to
        # the same function
        f_ast.name = self.inliner.fresh(f_ast.name)

        # If function has decorators, deal with those first. Just inline decorator call
        # and stop there.
        decorators = f_ast.decorator_list
        assert len(decorators) <= 1
        if len(decorators) == 1:
            d = decorators[0]
            builtin_decorator = (isinstance(d, ast.Name)
                                 and (d.id in ['property', 'classmethod']))
            derived_decorator = (isinstance(d, ast.Attribute)
                                 and (d.attr in ['setter']))
            if not (builtin_decorator or derived_decorator):
                self._expand_decorators(new_stmts, f_ast, call_expr, call_obj,
                                        ret_var)
                return new_stmts

        # If we're inlining a decorator, we need to remove @functools.wraps calls
        # to avoid messing up inspect.getsource
        for stmt in f_ast.body:
            RemoveFunctoolsWraps().visit(stmt)

        # If the function is a method (which we proxy by first arg being named "self"),
        # then we need to replace uses of special "self" and "super" keywords.
        args_def = f_ast.args
        if len(args_def.args) > 0 and args_def.args[0].arg == 'self':
            self._replace_self_super(f_ast, cls, call_expr)

        # Add bindings from arguments in the call expression to arguments in function def
        self._bind_arguments(f_ast, call_expr, new_stmts)

        # Add an explicit return None at the end to reify implicit return
        f_ast.body.append(parse_stmt("return None"))

        # Iteratively replace all return statements with conditional assignments to
        # the ret_var. See ReplaceReturn in visitors.py for how this works.
        while True:
            replacer = ReplaceReturn(ret_var)
            replacer.visit(f_ast)
            if not replacer.found_return:
                break

        # Inline function body
        new_stmts.extend(f_ast.body)

        # If we're inlining a function not defined in the top-level source, then
        # add imports for all the nonlocal (global + closure) variables
        if not self.inliner.is_source_obj(call_obj):
            used = self._get_nonlocal_vars(f_ast, call_obj)
            file_imports = collect_imports(call_obj)
            for name, value in used.items():
                if name == 'self':  # HACK
                    continue
                imprt = self.generate_imports(name,
                                              value,
                                              call_obj=call_obj,
                                              file_imports=file_imports)
                if imprt is not None:
                    new_stmts.insert(0, imprt)

        return new_stmts

    def generate_imports(self, name, globl, call_obj, file_imports):
        """
        Generate an import statement for a (name, runtime object) pair.


        """
        if name in file_imports:
            return file_imports[name]

        # If we're importing a module, then add an import directly
        if inspect.ismodule(globl):
            # Add an alias if the imported name is different from the module name
            alias = name if globl.__name__ != name else None
            return ast.Import([ast.alias(name=globl.__name__, asname=alias)])
        else:
            # Get module where global is defined
            mod = inspect.getmodule(globl)

            # TODO: When is mod None?
            if mod is None or mod is typing:
                try:
                    mod_value = obj_to_ast(globl)
                    return ast.Assign(targets=[make_name(name)],
                                      value=mod_value)
                except ObjConversionException:
                    return ast.ImportFrom(
                        module=inspect.getmodule(call_obj).__name__,
                        names=[ast.alias(name=name, asname=None)],
                        level=0)

            # Can't import builtins
            elif mod == __builtins__:
                return None

            # If the value is a class or function, then import it from the defining
            # module
            elif inspect.isclass(globl) or inspect.isfunction(globl):
                return ast.ImportFrom(module=mod.__name__,
                                      names=[ast.alias(name=name, asname=None)],
                                      level=0)

            # Otherwise import it from the module using the global
            elif call_obj is not None:
                return ast.ImportFrom(
                    module=inspect.getmodule(call_obj).__name__,
                    names=[ast.alias(name=name, asname=None)],
                    level=0)

    def expand_comprehension(self, comp, ret_var, call_finder):
        forloop = None
        for gen in reversed(comp.generators):
            if forloop is None:
                body = []
                body.extend(
                    self.inline_function(call_finder.call_obj,
                                         call_finder.call_expr,
                                         call_finder.ret_var, self.globls))
                append = parse_expr(f'{ret_var}.append()')
                append.args = [comp.elt]
                body.append(ast.Expr(append))
            else:
                body = [forloop]

            forloop = ast.For(target=gen.target,
                              iter=gen.iter,
                              body=body,
                              orelse=[])

        init = parse_stmt(f'{ret_var} = []')

        return [init, forloop]

    def expand_ifexp(self, ifexp, ret_var):
        def assgn(value):
            return ast.Assign(targets=[make_name(ret_var)], value=value)

        return [
            ast.If(test=ifexp.test,
                   body=[assgn(ifexp.body)],
                   orelse=[assgn(ifexp.orelse)])
        ]

    def expand_with(self, withstmt):
        assert len(withstmt.items) == 1
        item = withstmt.items[0]

        if item.optional_vars is not None:
            assert isinstance(item.optional_vars, ast.Name)
            name = item.optional_vars.id
        else:
            name = self.inliner.fresh('withctx')

        init = ast.Assign(targets=[make_name(name)], value=item.context_expr)
        enter = parse_stmt(f'{name}.__enter__()')
        exit_ = parse_stmt(f'{name}.__exit__()')
        return [init, enter] + withstmt.body + [exit_]



from setuptools import setup, find_packages

if __name__ == "__main__":
    setup(name='inliner',
          version='0.3.1',
          description='Human-readable inlining of Python code',
          url='http://github.com/willcrichton/inliner',
          author='Will Crichton',
          author_email='wcrichto@cs.stanford.edu',
          license='Apache 2.0',
          packages=find_packages(),
          install_requires=[
              'astor', 'iterextras', 'numpy', 'pandas', 'astpretty'
          ],
          data_files=[('share/jupyter/nbextensions/inliner', [
              'inliner_jupyter/dist/notebook.js',
              'inliner_jupyter/dist/notebook.js.map',
          ]),
                      ('etc/jupyter/nbconfig/notebook.d',
                       ['inliner_jupyter/inliner.json'])],
          setup_requires=['pytest-runner'],
          tests_require=['pytest', 'seaborn'],
          zip_safe=False)



import ast
import copy
import inspect

from .common import *


class Rename(ast.NodeTransformer):
    def __init__(self, src, dst):
        self.src = src
        self.dst = dst

    def visit_FunctionDef(self, fdef):
        if fdef.name == self.src:
            fdef.name = self.dst
            self.generic_visit(fdef)
        else:
            args = fdef.args
            arg_names = set()

            for arg in args.args:
                arg_names.add(arg.arg)
            if args.vararg is not None:
                arg_names.add(args.vararg.arg)
            if args.kwarg is not None:
                arg_names.add(args.kwarg.arg)

            if self.src not in arg_names:
                self.generic_visit(fdef)
        return fdef

    def visit_Name(self, name):
        if name.id == self.src:
            name.id = self.dst
        return name


class Replace(ast.NodeTransformer):
    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.replaced = False

    def visit_Name(self, name):
        if name.id == self.name:
            self.replaced = True
            return self.value
        else:
            return name


class FindAssignments(ast.NodeVisitor):
    def __init__(self):
        self.names = set()

    def visit_Assign(self, assgn):
        for t in assgn.targets:
            if isinstance(t, ast.Name):
                self.names.add(t.id)


class ReplaceReturn(ast.NodeTransformer):
    def __init__(self, name):
        self.name = name
        self.toplevel = True
        self.found_return = False
        self.if_wrapper = parse_stmt(
            textwrap.dedent("""
        if "{name}" not in locals() and "{name}" not in globals():
            pass
            """.format(name=self.name)))

    def visit_Return(self, stmt):
        if_stmt = copy.deepcopy(self.if_wrapper)
        # A naked return (without a value) will have stmt.value = None
        value = stmt.value if stmt.value is not None else ast.NameConstant(None)
        if_stmt.body[0] = ast.Assign(targets=[make_name(self.name)],
                                     value=value)
        self.found_return = True
        return if_stmt

    def visit_FunctionDef(self, fdef):
        # no recurse to avoid messing up inline functions
        if self.toplevel:
            self.toplevel = False
            self.generic_visit(fdef)
        return fdef

    def generic_visit(self, node):
        for field, old_value in ast.iter_fields(node):
            if isinstance(old_value, list):
                new_values = []
                for i, cur_value in enumerate(old_value):
                    if isinstance(cur_value, ast.AST):
                        value = self.visit(cur_value)

                        stmt_types = (ast.For, ast.If, ast.With,
                                      ast.FunctionDef, ast.Assign, ast.While)
                        if isinstance(node, stmt_types) and self.found_return:
                            new_values.append(value)
                            if i < len(old_value) - 1:
                                if_stmt = copy.deepcopy(self.if_wrapper)
                                if_stmt.body = old_value[i + 1:]
                                new_values.append(if_stmt)
                            break

                        if value is None:
                            continue
                        elif not isinstance(value, ast.AST):
                            new_values.extend(value)
                            continue

                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)

        return node


class ReplaceYield(ast.NodeTransformer):
    def __init__(self, ret_var):
        self.ret_var = ret_var

    def visit_Yield(self, expr):
        append = parse_expr(f'{self.ret_var}.append()')
        append.args.append(expr.value)
        return append


class ReplaceSelf(ast.NodeTransformer):
    def __init__(self, cls, globls):
        self.cls = cls
        self.globls = globls

    def visit_Call(self, expr):
        if isinstance(expr.func, ast.Attribute) and \
            isinstance(expr.func.value, ast.Name) and \
            expr.func.value.id == 'self':

            if hasattr(self.cls, expr.func.attr):
                attr = getattr(self.cls, expr.func.attr)

                if inspect.isfunction(attr):
                    expr.func.value = make_name(self.cls.__name__)
                    expr.args.insert(0, make_name('self'))

        return expr


class ReplaceSuper(ast.NodeTransformer):
    def __init__(self, cls):
        self.cls = cls

    def visit_Call(self, call):
        if isinstance(call.func, ast.Attribute) and \
           isinstance(call.func.value, ast.Call) and \
           isinstance(call.func.value.func, ast.Name) and \
           call.func.value.func.id == 'super':
            call.func.value = make_name(self.cls.__name__)
            call.args.insert(0, make_name('self'))
        self.generic_visit(call)
        return call


class UsedGlobals(ast.NodeVisitor):
    def __init__(self, globls):
        self.globls = globls
        self.used = {}

    def visit_Name(self, name):
        if name.id in self.globls:
            self.used[name.id] = self.globls[name.id]


class CollectLineNumbers(ast.NodeVisitor):
    def __init__(self):
        self.linenos = set()

    def generic_visit(self, node):
        if hasattr(node, 'lineno'):
            self.linenos.add(node.lineno)
        super().generic_visit(node)


class CollectImports(ast.NodeVisitor):
    def __init__(self, mod):
        self.imprts = {}
        self.mod = mod

    def visit_Import(self, imprt):
        for alias in imprt.names:
            name = alias.asname if alias.asname is not None else alias.name
            self.imprts[name] = ast.Import(names=[alias])

    def visit_ImportFrom(self, imprt):
        for alias in imprt.names:
            name = alias.asname if alias.asname is not None else alias.name

            if imprt.level > 0:
                parts = self.mod.split('.')
                mod_level = '.'.join(
                    parts[:-imprt.level]) if len(parts) > 1 else parts[0]
                if imprt.module is not None:
                    module = f'{mod_level}.{imprt.module}'
                else:
                    module = mod_level
            else:
                module = imprt.module

            self.imprts[name] = ast.ImportFrom(module=module,
                                               names=[alias],
                                               level=0)


def collect_imports(call_obj):
    import_collector = CollectImports(mod=inspect.getmodule(call_obj).__name__)
    import_collector.visit(
        ast.parse(open(inspect.getsourcefile(call_obj)).read()))
    return import_collector.imprts


class RemoveFunctoolsWraps(ast.NodeTransformer):
    def visit_FunctionDef(self, fdef):
        if len(fdef.decorator_list) == 1:
            dec = fdef.decorator_list[0]
            if isinstance(dec, ast.Call) and compare_ast(
                    dec.func, parse_expr("functools.wraps")):
                fdef.decorator_list = []
        return fdef

    def visit_Call(self, call):
        if compare_ast(call.func, parse_expr("update_wrapper")):
            return call.args[0]
        self.generic_visit(call)
        return call


class CollectModules(ast.NodeVisitor):
    def __init__(self, globls):
        self.globls = globls
        self.modules = {}

    def generic_visit(self, node):
        if isinstance(node, (ast.Name, ast.Attribute)):
            try:
                src = a2s(node)
                obj = eval(src, self.globls, self.globls)
            except Exception:
                obj = None

            if obj is not None:
                mod = inspect.getmodule(obj)
                if mod is not None:
                    name = mod.__name__
                    if name not in self.modules:
                        self.modules[name] = src
                return

        super().generic_visit(node)



import ast
from collections import defaultdict

from ..tracer import Tracer


class RemoveEmptyBlocks(ast.NodeTransformer):
    def generic_visit(self, node):
        super().generic_visit(node)

        if hasattr(node, 'body') and \
           ((isinstance(node.body, list) and len(node.body) == 0) or \
            node.body is None):
            return None
        return node


class BasePass(ast.NodeTransformer):
    tracer_args = None

    def __init__(self, inliner):
        self.inliner = inliner

        if self.tracer_args is not None:
            if inliner._tracer_cache is not None and \
               self.tracer_args == inliner._tracer_cache[0]:
                tracer = inliner._tracer_cache[1]
            else:
                prog = inliner.make_program()
                tracer = Tracer(prog, inliner.globls, **self.tracer_args)
                tracer.trace()
                inliner.module = ast.parse(prog)

            self.tracer = tracer
            self.globls = tracer.globls

        self.change = False

    def after_visit(self, mod):
        pass

    def visit_Module(self, mod):
        self.generic_visit(mod)
        RemoveEmptyBlocks().visit(mod)
        self.after_visit(mod)
        return mod

    def visit_FunctionDef(self, fdef):
        # Don't recurse into inline function definitions
        return fdef

    def run(self, **kwargs):
        self.args = defaultdict(lambda: None)
        for k, v in kwargs.items():
            self.args[k] = v

        self.visit(self.inliner.module)

        return self.change



import astor
from astor.code_gen import SourceGenerator
import textwrap
import itertools
import ast
import re
from iterextras import unzip
import typing
import math
from astpretty import pprint as pprintast
import copy
import inspect

SEP = '___'
COMMENT_MARKER = '__comment: '


def parse_stmt(s):
    return ast.parse(textwrap.dedent(s)).body[0]


def parse_expr(s):
    return parse_stmt(textwrap.dedent(s)).value


class SourceGeneratorWithComments(SourceGenerator):
    COMMENTS = False

    def visit_Str(self, node):
        if self.__class__.COMMENTS and node.s.startswith(COMMENT_MARKER):
            s = node.s[len(COMMENT_MARKER):]
            indent = self.indent_with * self.indentation
            comment = '\n'.join([f'{indent}# {part}' for part in s.split('\n')])
            self.write('#\n' + comment)
        else:
            super().visit_Str(node)


def a2s(a, comments=False):
    SourceGeneratorWithComments.COMMENTS = comments
    outp = astor.to_source(a,
                           source_generator_class=SourceGeneratorWithComments)
    return re.sub(r'^\s*#\s*\n', '\n', outp, flags=re.MULTILINE).rstrip()


# https://stackoverflow.com/questions/3312989/elegant-way-to-test-python-asts-for-equality-not-reference-or-object-identity
def compare_ast(node1, node2):
    if type(node1) is not type(node2):
        return False
    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in ('lineno', 'col_offset', 'ctx', '_pp'):
                continue
            if not compare_ast(v, getattr(node2, k)):
                return False
        return True
    elif isinstance(node1, list):
        return all(itertools.starmap(compare_ast, zip(node1, node2)))
    else:
        return node1 == node2


class ObjConversionException(Exception):
    pass


def obj_to_ast(obj):
    if isinstance(obj, tuple):
        return ast.Tuple(elts=tuple(map(obj_to_ast, obj)))
    elif isinstance(obj, dict):
        k, v = unzip([(obj_to_ast(k), obj_to_ast(v)) for k, v in obj.items()])
        return ast.Dict(k, v)
    elif isinstance(obj, list):
        return ast.List(list(map(obj_to_ast, obj)))
    elif isinstance(obj, type):
        return ast.Name(id=obj.__name__)
    elif isinstance(obj, int):
        return ast.Num(obj)
    elif isinstance(obj, str):
        return ast.Str(obj)
    elif obj is None:
        return ast.NameConstant(None)
    elif isinstance(obj, (typing._GenericAlias, typing._SpecialForm)):
        # TODO: types
        # issue was in pandas, where importing pandas._typing.Axis would
        # resolve module to typing, attempt to do "from typing import Axis"
        return ast.NameConstant(None)
    elif isinstance(obj, float) and math.isinf(obj):
        return parse_expr('float("inf")')
    elif isinstance(obj, bytes):
        return ast.Bytes(s=obj)
    else:
        raise ObjConversionException(f"No converter for {obj}")


def can_convert_obj_to_ast(obj):
    try:
        obj_to_ast(obj)
        return True
    except ObjConversionException:
        return False


def robust_eq(obj1, obj2):
    import pandas as pd
    import numpy as np

    if type(obj1) != type(obj2):
        return False
    elif isinstance(obj1, pd.DataFrame) or isinstance(obj1, pd.Series):
        return obj1.equals(obj2)
    elif isinstance(obj1, np.ndarray):
        return np.array_equal(obj1, obj2)
    elif isinstance(obj1, tuple) or isinstance(obj1, list):
        return len(obj1) == len(obj2) and all(
            map(lambda t: robust_eq(*t), zip(obj1, obj2)))
    elif isinstance(obj1, (int, float, str)) or \
         inspect.isfunction(obj1) or \
         inspect.isclass(obj1) or \
         obj1 is None:
        return obj1 == obj2

    else:
        return False


def make_name(s):
    return ast.Name(id=s, ctx=ast.Load())


def try_copy(v):
    try:
        return copy.deepcopy(v)
    except Exception:
        try:
            return copy.copy(v)
        except Exception:
            return v


def get_function_locals(f):
    if f.__closure__ is not None and len(f.__closure__) > 0:
        return {
            var: cell.cell_contents
            for var, cell in zip(f.__code__.co_freevars, f.__closure__)
        }

    return {}


class IsEffectFree(ast.NodeVisitor):
    def __init__(self):
        self.effect_free = True

        # HACK: how can we actually detect purity?
        func_whitelist = ['str', 'utils.to_utf8', 'LooseVersion']
        self.func_whitelist = [parse_expr(s) for s in func_whitelist]

        self.ast_whitelist = (ast.Num, ast.Str, ast.Name, ast.NameConstant,
                              ast.Load, ast.List, ast.Bytes, ast.Tuple, ast.Set,
                              ast.Dict, ast.Attribute)

    def generic_visit(self, node):
        if isinstance(node, ast.FunctionDef):
            return

        if isinstance(node, ast.Call):
            if not any([compare_ast(node.func, f)
                        for f in self.func_whitelist]):
                self.effect_free = False
        else:
            if not isinstance(node, self.ast_whitelist):
                self.effect_free = False

        super().generic_visit(node)


def is_effect_free(node):
    visitor = IsEffectFree()
    visitor.visit(node)
    return visitor.effect_free


class TreeSize(ast.NodeVisitor):
    def __init__(self):
        self.size = 0

    def generic_visit(self, node):
        self.size += 1
        super().generic_visit(node)


def tree_size(node):
    visitor = TreeSize()
    visitor.visit(node)
    return visitor.size
